from dataclasses import field
import logging
from typing import ClassVar, Mapping
from pydantic.dataclasses import dataclass


@dataclass
class Plugin:
    TYPE: ClassVar[str] = "message_aggregate"
    by: list
    output_fields: list
    prioritize: str = "last"
    _aggregation_context: dict = field(default_factory=dict)

    def __post_init__(self):
        # can be implemented with pydantic in a cleaner way
        assert self.prioritize in {
            "last",
            "first",
        }, "Invalid priorization method"

    def callback(self, message: Mapping, context: Mapping):
        message = dict(message)
        merge_key = []

        for merge_by_field in self.by:
            if merge_by_field not in message:
                return
            merge_key.append(message[merge_by_field])

        merge_key = tuple(merge_key)
        if merge_key not in self._aggregation_context:
            self._aggregation_context[merge_key] = message
            return

        first_message = self._aggregation_context.pop(merge_key)
        last_message = message
        output_message = {}

        for output_field in self.output_fields:
            in_first = output_field in first_message
            in_last = output_field in last_message

            match (in_first, in_last, self.prioritize):
                case (True, False, _) | (True, True, "first"):
                    field_value = first_message[output_field]
                case (False, True, _) | (True, True, "last"):
                    field_value = last_message[output_field]
                case (False, False, _):
                    logging.error(
                        f"Missing output field {output_field} in both messages, "
                        f"unable to aggregate first message: {first_message} and last message: {last_message}"
                    )
                    return
                case _:
                    logging.error(
                        f"Well thats unexpected, you should probly take a look at this {(in_first, in_last, self.prioritize) = }"
                    )
                    return

            output_message[output_field] = field_value

        return output_message



import logging
from typing import Any, Mapping, MutableMapping

from forum.plugin.node.node import ContextBuilder, ContextDestroyer, Node
from forum.topic import TOPIC


class NodeCallbackAdapter:
    def __init__(self, name: str, node: Node):
        self.name = name
        self.input_topic = f"{name}.input"
        self.output_topic = f"{name}.output"
        self.callback = node.callback
        self.context = None

        if isinstance(node, ContextBuilder):
            self.context = node.build_context(name)

    def __call__(self, message: Mapping):
        response = self.callback(message, self.context)
        if response is not None:
            TOPIC.publish(self.output_topic, response)

    def __repr__(self) -> str:
        return f"NodeCallbackAdapter(name={self.name})"


class NodeManager:
    def __init__(self) -> None:
        self.nodes: MutableMapping[str, Node] = {}
        self.nodes_context: MutableMapping[str, Any] = {}

    def add_node(self, name: str, node: Node):
        self.nodes[name] = node
        node_callback_adapter = NodeCallbackAdapter(name, node)
        if node_callback_adapter is not None:
            self.nodes_context[name] = node_callback_adapter.context
        TOPIC.subscribe(
            node_callback_adapter.input_topic,
            node_callback_adapter,
        )

    def remove_node(self, name: str):
        try:
            node = self.nodes[name]
        except KeyError as key_error:
            raise RuntimeError(
                f"Failed to remove node {name}, node not found"
            ) from key_error

        if not isinstance(node, ContextDestroyer):
            return

        try:
            node_context = self.nodes_context[name]
            node.destroy_context(**node_context)
        except KeyError as key_error:
            raise RuntimeError(
                f"Failed to destroy context for {name}, node context not found"
            ) from key_error



"""Factory for creating a registered."""
from __future__ import annotations

import logging
from typing import Callable, ClassVar, MutableMapping
from dataclasses import dataclass, field
from forum.plugin import Plugin
from forum.plugin.node.node import Node


@dataclass(slots=True)
class NodeFactory:
    _creation_funcs: ClassVar[MutableMapping[str, Callable]] = {}

    def register(self, type_name: str, creator_fn: Callable) -> None:
        """Register a new registered type."""
        self._creation_funcs[type_name] = creator_fn

    def unregister(self, type_name: str) -> None:
        """Unregister a registered type."""
        self._creation_funcs.pop(type_name, None)

    def create(self, type_name: str, **kwargs) -> Node:
        """Create a registered of a specific type"""
        try:
            creator_func = self._creation_funcs[type_name]
        except KeyError:
            raise ValueError(f"Unknown registered type {type_name!r}") from None
        logging.info("Exec creator function: %s(%s)", type_name, kwargs)
        return creator_func(**kwargs)



"""A simple plugin loader.
from https://github.com/ArjanCodes/2021-plugin-architecture/blob/main/after/game/loader.py
"""
from __future__ import annotations

import importlib
import logging
from types import ModuleType
from typing import Mapping, Protocol

from forum.plugin.node import Node, NodeFactory
from forum.plugin import Plugin, PluginFactory, PluginModule

FACTORIES_MAPPING: Mapping[Plugin, PluginFactory] = {Node: NodeFactory()}


class PluginRegisterer:
    interface: Protocol
    factory: object


def import_module(name: str) -> ModuleType | None:
    """Imports a module given a name."""
    try:
        module = importlib.import_module(name)  # type: ignore
    except (ModuleNotFoundError, ImportError) as exeption:
        logging.error(f"Failed to import module {name}, {exeption}")
        module = None
    return module


def register_plugin(module: PluginModule):
    for protocol, factory in FACTORIES_MAPPING.items():
        plugin = module.Plugin
        if isinstance(plugin, protocol):  # type: ignore
            return factory.register(plugin.TYPE, plugin)
    logging.error(
        "Unsuported plugin type, must confome with one of the following "
        f"protocols: {FACTORIES_MAPPING.keys()}"
    )


def load_plugins(plugins: list[str]) -> Mapping[Plugin, PluginFactory]:
    """Loads the plugins defined in the plugins list."""
    n_plugins = len(plugins)
    logging.info(f"🔌 Plugins found: {n_plugins}")

    for plugin_file in plugins:
        module = import_module(plugin_file)
        if not isinstance(module, PluginModule):
            logging.info(
                f"❌ Failed to load {plugin_file} plugin, it does not respect the PluginModule protocol"
            )
            continue
        register_plugin(module)
        logging.info(f"✔️ Loaded {plugin_file} plugin")

    return dict(FACTORIES_MAPPING)



from __future__ import annotations

from dataclasses import field
from pathlib import Path
from typing import Any
from pydantic import BaseModel
from forum.plugin.node import Node

import yaml


# @dataclass
class Config(BaseModel):
    # Warning this are not class atributes and dont use
    # a default_factory because of pydantict
    # please see https://stackoverflow.com/questions/63793662/how-to-give-a-pydantic-list-field-a-default-value
    plugins: list[str] = []
    nodes: dict[str, dict[str, Any]] = {}
    topics: dict[str, str | list[str]] = {}

    @classmethod
    def from_yaml(cls, yaml_location: str):
        config_location = Path(yaml_location)

        with config_location.open(encoding="utf-8") as config_file:
            config_data = yaml.safe_load(config_file)

        return cls(**config_data)



from functools import partial
import logging
from threading import Thread
from typing import ClassVar, Mapping

import telebot
from pydantic import BaseModel
from pydantic.dataclasses import dataclass
from telebot.types import Message as TelegramMessage

from forum.topic import TOPIC
from forum.nodes.chat_message import ChatMessage


@dataclass
class TelegramBotAPI:
    def __init__(self, instance_name: str, api_key: str):
        self._instance_name = instance_name
        self.bot = telebot.TeleBot(api_key)

        handlers_output_topic = f"{instance_name}.output"
        chat_start_handler = partial(self.chat_start_handler, handlers_output_topic)
        self.bot.register_message_handler(
            chat_start_handler, commands=["help", "start"]
        )

        chat_text_handler = partial(self.chat_text_handler, handlers_output_topic)
        self.bot.register_message_handler(chat_text_handler, content_types=["text"])

        audio_message_handler = partial(
            self.audio_message_handler, handlers_output_topic
        )
        self.bot.register_message_handler(
            audio_message_handler, content_types=["voice"]
        )
        self.pooling_thread = Thread(
            name="telebot_pooling",
            target=self.bot.infinity_polling,
        )

    @staticmethod
    def chat_start_handler(output_topic: str, telegram_message: TelegramMessage):
        greeting_text = "Bom dia, tudo bem?"
        text_message = ChatMessage(
            chat_id=str(telegram_message.chat.id),
            input_message_id=str(telegram_message.id),
            type="text",
            content=greeting_text,
        )

        TOPIC.publish(output_topic, text_message.dict())

    @staticmethod
    def chat_text_handler(output_topic: str, telegram_message: TelegramMessage):
        logging.info("TEXT RECIVED")
        if not telegram_message.text:
            return

        text_message = ChatMessage(
            chat_id=str(telegram_message.chat.id),
            input_message_id=str(telegram_message.id),
            type="text",
            content=telegram_message.text,
        )

        TOPIC.publish(output_topic, text_message.dict())

    @staticmethod
    def audio_message_handler(output_topic: str, telegram_message: TelegramMessage):
        response_text = "Amo esse som!"
        text_message = ChatMessage(
            chat_id=str(telegram_message.chat.id),
            input_message_id=str(telegram_message.id),
            type="text",
            content=response_text,
        )

        TOPIC.publish(output_topic, text_message.dict())

    def send_message(self, chat_message: ChatMessage):
        if chat_message.type == "text":
            chat_id = int(chat_message.chat_id)
            self.bot.send_message(
                chat_id,
                chat_message.content,
            )
        else:
            raise RuntimeError("Invalid message type")

    def start(self):
        self.pooling_thread.start()

    def shutdown(self, timeout_seconds: float = 5):
        self.pooling_thread.join(timeout=timeout_seconds)


class TelegramContext(BaseModel):
    telegram_api: TelegramBotAPI


@dataclass
class Plugin:
    TYPE: ClassVar[str] = "telegram_bot"
    api_key: str

    def build_context(self, instance_name: str) -> Mapping:
        logging.info("telegram context created")
        telegram_api = TelegramBotAPI(instance_name, self.api_key)

        telegram_api.start()
        telegram_context = TelegramContext(telegram_api=telegram_api)
        return telegram_context.dict()

    def callback(self, message: Mapping, context: Mapping):
        telegram_context = TelegramContext(**context)
        telegram_api = telegram_context.telegram_api
        chat_message = ChatMessage(**message)
        telegram_api.send_message(chat_message)

    def destroy_context(self, context: Mapping):
        telegram_context = TelegramContext(**context)
        telegram_context.telegram_api.shutdown()



from copy import copy
from queue import Queue
from threading import Lock, Thread
from typing import Any, Callable, Mapping, MutableMapping
from dataclasses import dataclass
from venv import logger


class CallableList(list[Callable]):
    def __call__(self, *args: Any, **kwargs: Any):
        for item in self:
            item(*args, **kwargs)


SubscriberMap = MutableMapping[str, CallableList]


class PubSub:
    def __init__(self):
        # Locked added as an precaution for more information visit
        # https://github.com/google/styleguide/blob/91d6e367e384b0d8aaaf7ce95029514fcdf38651/pyguide.md#218-threading
        self._subscribers_lock = Lock()
        self.subscribers: SubscriberMap = {}
        self.queue: Queue = Queue()
        self.message_processing_thread = Thread(target=self._process_messages)

    def subscribe(self, subscription_topic: str, callback: Callable):
        with self._subscribers_lock:
            if subscription_topic not in self.subscribers:
                self.subscribers[subscription_topic] = CallableList()

            self.subscribers[subscription_topic].append(callback)

    def publish(self, destination_topic: str, message: Mapping | None):
        if message is not None:
            self.queue.put((destination_topic, message))

    def _process_messages(self):
        while True:
            destination_topic, message = self.queue.get()
            logger.info(f"[{destination_topic}]: {message}")

            if not destination_topic in self.subscribers:
                continue

            with self._subscribers_lock:
                # A shallow copy is made so changes in this subscribers
                # CallableList wont cause trouble this must be refectored
                # to achive better perfomance in topic comunication since
                # it wont scale well
                topic_callbacks = copy(self.subscribers[destination_topic])

            topic_callbacks(message)

    def start(self):
        self.message_processing_thread.start()

    def shutdown(self):
        self.queue.join()
        self.message_processing_thread.join()


@dataclass
class TopicFowarder:
    output_topic: str

    def __call__(self, message: Mapping) -> Any:
        TOPIC.publish(self.output_topic, message)


TOPIC = PubSub()
TOPIC.start()



from typing import ClassVar, Mapping, Protocol, runtime_checkable


@runtime_checkable
class Node(Protocol):
    TYPE: ClassVar[str]

    def callback(
        self, message: Mapping, context: Mapping | None = None
    ) -> Mapping | None:
        ...

    def __call__(self, *args, **kwargs):
        ...


@runtime_checkable
class ContextBuilder(Protocol):
    def build_context(self, instance_name: str) -> Mapping | None:
        ...


@runtime_checkable
class ContextDestroyer(Protocol):
    def destroy_context(self, context: Mapping) -> None:
        ...



from typing import Callable, Protocol, runtime_checkable

from forum.plugin.protocol import Plugin


@runtime_checkable
class PluginFactory(Protocol):
    def register(self, type_name: str, creator_fn: Callable) -> None:
        """Unregisters a register type."""
        ...

    def unregister(self, type_name: str) -> None:
        """Unregisters a registered type."""
        ...

    def create(self, type_name: str, **kwargs) -> Plugin:
        """Creates object of given type."""
        ...



from dataclasses import dataclass
from inspect import signature
from typing import ClassVar, Mapping, Protocol, runtime_checkable


@runtime_checkable
class Node(Protocol):
    TYPE: ClassVar[str]

    def callback(
        self, message: Mapping, context: Mapping | None = None
    ) -> Mapping | None:
        ...

    def __call__(self, *args, **kwargs):
        ...


@runtime_checkable
class InstanceAware(Protocol):
    def __init__(self, instance_name: str, *args, **kwargs) -> None:
        ...


@dataclass
class Plugin:
    TYPE: ClassVar[str] = "telegram_bot"
    api_key: str

    def callback(self, message: Mapping, context: Mapping):
        pass


parameter = signature(Plugin.__init__).parameters.keys()
print("is node:", isinstance(Plugin, Node))
print("is instance aware:", issubclass(Plugin, InstanceAware))



from typing import ClassVar, Mapping
from forum.nodes.chat_message import ChatMessage
from pydantic.dataclasses import dataclass


# this can be waaaaay more generic
# plz refactor
@dataclass
class Plugin:
    TYPE: ClassVar[str] = "message_filter"
    content_type: str | None = None

    def callback(self, message: Mapping, context: Mapping):
        chat_message = ChatMessage(**message)

        if self.content_type and self.content_type != chat_message.type:
            return

        return chat_message.dict()



import json
from pathlib import Path
from typing import ClassVar, Mapping
from pydantic.dataclasses import dataclass

import requests
from pydantic import BaseModel


class STTAPI(BaseModel):
    address: str
    port: int
    language: str

    def request_transcription(self, file_location: str | Path) -> str | None:
        endpoint = f"http://{self.address}:{self.port}/stt"
        audio_file_location = Path(file_location)
        stt_request = {"language": "pt"}

        with open(file_location, "rb") as speach_file:
            response = requests.post(
                url=endpoint,
                data={"stt_request": json.dumps(stt_request)},
                files={
                    "audio_file": (audio_file_location.name, speach_file, "audio/wav")
                },
                timeout=30,
            )

        response = response.json()
        response["text"] = response["text"].strip().lower()
        return response["text"]


class STTContext(BaseModel):
    stt_api: STTAPI


class STTMessage(BaseModel):
    file_location: str


class STTResponse(BaseModel):
    text: str | None


@dataclass
class Plugin:
    TYPE: ClassVar[str] = "stt"
    address: str
    port: int
    language: str

    def build_context(self, instance_name: str) -> Mapping:
        stt_api = STTAPI(address=self.address, port=self.port, language=self.language)
        stt_context = STTContext(stt_api=stt_api)
        return stt_context.dict()

    def callback(self, message: Mapping, context: Mapping):
        stt_message = STTMessage(**message)
        stt_contex = STTContext(**context)
        stt_api = stt_contex.stt_api
        response_data = stt_api.request_transcription(stt_message.file_location)
        stt_response = STTResponse(text=response_data)

        return stt_response.dict()



from __future__ import annotations
from langchain.vectorstores import Chroma

from typing import ClassVar, Mapping
from pydantic.dataclasses import dataclass
from pydantic import BaseModel
from langchain.llms import LlamaCpp
from langchain.memory import ConversationSummaryBufferMemory
from langchain import LLMChain, PromptTemplate
from operator import itemgetter
from langchain.chains import (
    ConversationalRetrievalChain,
    SimpleSequentialChain,
    VectorDBQA,
)
from langchain.prompts import ChatPromptTemplate
from langchain.output_parsers import StructuredOutputParser


# class LLM:
#     def initialize():
#         # Create the retriever
#         vectorstore = Chroma(
#             collection_name="agent-memory", persist_directory=".persitence"
#         )
#         retriever = vectorstore.as_retriever()
#         template = """Answer the question based only on the following context:
#         {context}

#         Question: {question}

#         Answer in the following language: {language}
#         """
#         prompt = ChatPromptTemplate.from_template(template)

#         chain = (
#             {
#                 "context": itemgetter("question") | retriever,
#                 "question": itemgetter("question"),
#                 "language": itemgetter("language"),
#             }
#             | prompt
#             | model
#             | StructuredOutputParser()
# )


class LLMContext(BaseModel):
    inference_chain: LLMChain

    class Config:
        arbitrary_types_allowed = True


# probably not needded
class LLMMessage(BaseModel):
    text: str
    message_id: str


@dataclass
class Plugin:
    TYPE: ClassVar[str] = "llm"
    model_name: str
    system_prompt: str = ""
    # Uses LLAMA 2 prompt format by default
    system_prompt_prefix: str = "<<SYS>>\n"
    system_prompt_postfix: str = "\n<</SYS>>\n\n"
    instruction_prefix: str = "[INST]"
    instruction_postfix: str = "[/INST]"
    model_temperature: float = 0.7
    context_size: int = 4096

    def _build_prompt_template(self) -> PromptTemplate:
        prompt_template = (
            "{instruction_prefix}"
            "{system_prompt_prefix}"
            "{system_prompt}"
            "{system_prompt_postfix}"
            "Chat History:\n\n"
            "{chat_history} \n\n"
            "Human: {user_input}"
            "{instruction_postfix}\nAI: "
        )
        prompt_template = prompt_template.format(
            instruction_prefix=self.instruction_prefix,
            system_prompt_prefix=self.system_prompt_prefix,
            system_prompt=self.system_prompt,
            system_prompt_postfix=self.system_prompt_postfix,
            instruction_postfix=self.instruction_postfix,
            chat_history="{chat_history}",
            user_input="{user_input}",
        )
        print(prompt_template)

        prompt = PromptTemplate(
            input_variables=["chat_history", "user_input"], template=prompt_template
        )
        return prompt

    def build_context(self, instance_name: str) -> Mapping:
        llm = LlamaCpp(
            model_path=self.model_name,
            n_ctx=self.context_size,
            n_gpu_layers=40,
            temperature=self.model_temperature,
        )  # type: ignore
        memory = ConversationSummaryBufferMemory(
            llm=llm,
            memory_key="chat_history",
        )
        prompt_template = self._build_prompt_template()
        llm_chain = LLMChain(
            llm=llm,
            prompt=prompt_template,
            memory=memory,
            verbose=True,
        )
        llm_context = {"inference_chain": llm_chain}
        return llm_context

    def callback(self, message: Mapping, context: Mapping):
        llm_context = LLMContext(**context)
        llm_message = LLMMessage(**message)

        model = llm_context.inference_chain
        llm_message.text = model.predict(user_input=llm_message.text)

        return llm_message.dict()



from __future__ import annotations

import time
from pathlib import Path
from typing import ClassVar, Mapping

import requests
from pydantic import BaseModel
from pydantic.dataclasses import dataclass


class TTSAPI(BaseModel):
    address: str
    port: int

    def request_tts(self, prompt: str) -> str:
        endpoint = f"http://{self.address}:{self.port}/tts"
        message = {"text": prompt}
        response = requests.post(endpoint, json=message, timeout=300)
        current_time = time.time()
        audio_file_location = Path(f"{current_time}_tts_response.wav").absolute()
        with audio_file_location.open("wb") as audio_file:
            audio_file.write(response.content)

        return str(audio_file_location)


class TTSContext(BaseModel):
    tts_api: TTSAPI


class TTSMessage(BaseModel):
    text: str


class TTSResponse(BaseModel):
    file_location: str


@dataclass
class Plugin:
    TYPE: ClassVar[str] = "tts"
    address: str
    port: int

    def build_context(self, instance_name: str) -> Mapping:
        tts_api = TTSAPI(address=self.address, port=self.port)
        tts_context = TTSContext(tts_api=tts_api)
        return tts_context.dict()

    def callback(self, message: Mapping, context: Mapping):
        tts_message = TTSMessage(**message)
        tts_contex = TTSContext(**context)
        tts_api = tts_contex.tts_api
        response_data = tts_api.request_tts(tts_message.text)
        tts_response = TTSResponse(file_location=response_data)

        return tts_response.dict()



from __future__ import annotations

import logging
import sys
import time
from typing import Any
from pathlib import Path
from shutil import copy
import os


from forum.topic import TOPIC, TopicFowarder
from forum.config import Config
from forum.plugin.node import NodeFactory, NodeManager
from forum.plugin import plugin_loader
from forum.plugin.node import Node
import forum

CONFIG_LOCATION = "~/.local/share/byte/config.yml"
CONFIG_TEMPLATE_NAME = 'config.yml.template'

def initialize_config(
    telegram_api_key,
    llm_location,
    io_language,
):
    config_location = Path(CONFIG_LOCATION).absolute()
    module_location = Path(forum.__file__).parent.absolute()
    config_template_location = module_location / CONFIG_TEMPLATE_NAME

    config_template = config_template_location.read_text()
    config_template = config_template.format(
        telegram_api_key=telegram_api_key,
        llm_location=llm_location,
        io_language=io_language,
    )
    if not config_location.is_file():
        config_location.parent.mkdir(parents=True,exist_ok=True)
        config_location.write_text(config_template)


def load_nodes(
    node_factory: NodeFactory, nodes_configs: dict[str, dict[str, Any]]
) -> list[tuple[str, Node]]:
    nodes = []
    for name, config in nodes_configs.items():
        logging.info(f"Initializing service: {name}({config})")
        node_type = config.pop("type")
        node = node_factory.create(node_type, **config)
        nodes.append((name, node))
    return nodes


def map_topics(topics_map: dict[str, str | list[str]]) -> None:
    for input_topic, output_topics in topics_map.items():
        if isinstance(output_topics, str):
            output_topics = [output_topics]

        for output_topic in output_topics:
            logging.info(f"Fowarding: {input_topic} -> {output_topic}")
            fowarder = TopicFowarder(output_topic)
            TOPIC.subscribe(input_topic, fowarder)


def main(config_location: str):
    logging.basicConfig(level=logging.INFO)

    config = Config.from_yaml(config_location)
    logging.info(f"Loaded config: {config}")

    plugin_facotories = plugin_loader.load_plugins(config.plugins)
    nodes_factory: NodeFactory = plugin_facotories[Node]  # type: ignore
    nodes = load_nodes(nodes_factory, config.nodes)
    node_manager = NodeManager()
    for name, node in nodes:
        node_manager.add_node(name, node)
    map_topics(config.topics)

    try:
        while True:
            time.sleep(0.5)
    except KeyboardInterrupt:
        logging.info("🚪 Exiting")
    finally:
        TOPIC.shutdown()
        sys.exit(0)

def get_config_file_location() -> Path:
    user_home = os.environ.get('HOME')
    user_home = Path(user_home)
    default_config_dir_location = user_home/'.byte'
    default_config_file_location = default_config_dir_location / 'config.yml'
    config_file_location = os.environ.get('BYTE_CONFIG', default_config_file_location)
    config_file_location = Path(config_file_location)
    return config_file_location

if __name__ == "__main__":
    # init_params = (
    #     'telegram_api_key',
    #     'llm_location',
    #     'io_language',
    # )
    # init_param_values = {}
    # for init_param in init_params:
    #     init_param_values[init_param] = input(f'Value for {init_param}:')

    # initialize_config(
    #     **init_param_values
    # )
    
    config_file_location = get_config_file_location()
    
    # if not config_file_location.is_file():
    #     config_location = Path(CONFIG_LOCATION).absolute()
    #     module_location = Path(byte.__file__).parent.absolute()
    #     config_template_location = module_location / CONFIG_TEMPLATE_NAME
    
    main('/home/red/repos/byte/config.yml')



import logging
from typing import ClassVar, Mapping
from forum.nodes.chat_message import ChatMessage
from pydantic.dataclasses import dataclass


@dataclass
class Plugin:
    TYPE: ClassVar[str] = "message_transform"
    rename_fields: dict | None = None
    delete_fields: list | None = None
    set_fields: dict | None = None
    copy_fields: dict | None = None
    output_fields: list | None = None

    def callback(self, message: Mapping, context: Mapping):
        message = dict(message)

        # Review what happens when expressed fields doesnt exit
        # for now nothing happens, this is not a good idea for observability
        # this might be a setting

        if self.rename_fields is not None:
            for original_field_name, new_field_name in self.rename_fields.items():
                # think about what should happend when original field doesent exist
                message[new_field_name] = message.pop(original_field_name)

        if self.delete_fields is not None:
            for field_name in self.delete_fields:
                message.pop(field_name)

        if self.set_fields is not None:
            for field_name, value in self.set_fields.items():
                message[field_name] = value

        if self.copy_fields is not None:
            for original_field_name, new_field_name in self.copy_fields.items():
                message[new_field_name] = message[original_field_name]

        if self.output_fields:
            output_message = {}
            for output_field in self.output_fields:
                if output_field not in message:
                    logging.error(f"Missing output field {output_field} in message")
                    return
                output_message[output_field] = message[output_field]
            message = output_message

        return message
